/********************************************************************
 * ccmain.c     20-FEB-1992            Y.Takeuchi (T.I.T.)          *
 *                                                                  *
 * original: cc.c on SUN4 by Y.Yasu (KEK)                           *
 * for SUN-SPARC SFVME116 K2917 K3922                               *
 *                                                                  *
 * 13-JUL-1992 Y.Takeuchi    delete "NCC", modify "splx" etc.       *
 * 30-NOV-1992 Y.Takeuchi    modify CCPRI, ccioctl(), timeout, etc. *
 *  5-MAR-1993 Y.Takeuchi    modify for unified UNIX camlib         *
 * 17-MAR-1993 Y.Takeuchi    modify for re-execution                *
 * 17-APR-1993 Y.Takeuchi    fix bug in klexe()                     *
 ********************************************************************/
#include <sys/param.h>
#include <sys/buf.h>
#include <sys/types.h>
#include <sys/user.h>
#include <sys/uio.h>
#include <machine/psl.h>
#include <sundev/mbvar.h>
#include <errno.h>
#include "cc_config.h"
#include "cc_common.h"
#include "cc.h"
#include "k2917.h"

#define CCUNIT(dev) (minor(dev))
#define CCHPRI (PZERO - 1)
#define CCLPRI (PZERO + 1)

extern int hz;

int ccprobe();
int cctimeout();

/******************************************************
 * Data Structure                                     *
 ******************************************************/
struct mb_device *ccdinfo[NCC];
struct mb_driver ccdriver = {ccprobe, 0, 0, 0, 0, 0,
	sizeof(struct K_REG)*CC_K_MAX_BRANCH, "cc", ccdinfo, 0, 0, MDR_DMA, 0};
struct cc_device {
    int status;		  /* current system status */
    int cc_busy;          /* true if the device is busy */
    int execution;        /* true if the system call is under processing */
    struct mb_device *md; /* current pointer of md */
    struct K_REG *k;      /* current pointer of registers of K2917 */
    u_short *ptr_list;	  /* current pointer to list_buffer */
    int len_list;	  /* actual length processed */
    u_short *ptr_udata_s; /* first pointer to user's buffer address */
    u_short *ptr_udata;	  /* current pointer to user's buffer address */
    int len_udata_t;	  /* total length of user's buffer */
    int len_udata;	  /* actual length of user's buffer */
    u_short *ptr_kdata;	  /* current pointer to kernel's buffer address */
    int len_kdata;	  /* actual length of kernel's buffer processed */
    int s_reg;            /* S-register of KEK-list processing */ 
    int t_reg;            /* T-register of KEK-list processing */ 
    int a_reg;            /* A-register of KEK-list processing */ 
    int event_count;      /* event counter for KEK-list processing */
    u_short camac_qx;     /* CAMAC status (CSR of K2917) */
    int branch_flag;      /* internal flag for KEK-list processing */
    int max_branch;       /* maximum number of K2917 (fixed at 1) */
    int cur_branch;       /* current branch of K2917 (fixed at 0) */        
    int cur_crate;        /* current CAMAC crate address of K3922 (0-7) */
    int interrupt;        /* interrupt source */
    u_short mode;         /* K2917 mode */
    u_short naf;          /* CAMAC NAF (N<<9)+(A<<5)+F */
    int len;              /* data length */
    int retlen;           /* transfered data length */
    struct uio *uio; 
    dev_t dev; 
    int klist;            /* flag for K2917 list processing */
} ccdevice[NCC];

/******************************************************
 * Internal routines for CAMAC list processing        *
 ******************************************************/
extern int cc_stop(),
	cc_read16(),  cc_read24(),  cc_read16_a(),  cc_read24_a(),
	cc_write16(), cc_write24(), cc_write16_a(), cc_write24_a(),
	cc_ndt(), cc_setcrate_a(), cc_setcrate(), cc_genz(),
	cc_genc(), cc_remi(), cc_seti(), cc_enbint(), cc_dsbint(),
	cc_qwait(),
	cc_qs16(), cc_qs24(), cc_qs16_a(), cc_qs24_a(),
	cc_as16(), cc_as24(), cc_as16_a(), cc_as24_a(),
	cc_igq16(), cc_igq24(), cc_igq16_a(), cc_igq24_a(),
	cc_move_as(), cc_move_at(), cc_move_sa(), cc_move_st(),
	cc_move_ta(), cc_move_ts(),
	cc_move32_a(), cc_move32_s(), cc_move32_t(), cc_move16_a_ptr(),
	cc_move32_a_ptr(), cc_move16_ptr(), cc_move32_ptr(),
	cc_move16_a_head(), cc_move16_head_a(), cc_move16_head(),
	cc_load_ptr(), cc_load_remain(), cc_load_timer(), 
	cc_load_count(), cc_load_qx(),
	cc_head(), cc_exchg_at(), cc_swap_a(),
	cc_bis32_at(), cc_bis32_ta(), cc_bis32_a(), cc_bis32_t(),
	cc_bic32_at(), cc_bic32_ta(), cc_bic32_a(), cc_bic32_t(),
	cc_add32_at(), cc_add32_ta(), cc_add32_a(), cc_add32_t(),
	cc_sub32_at(), cc_sub32_ta(), cc_sub32_a(), cc_sub32_t(),
	cc_mul32_at(), cc_mul32_ta(), cc_mul32_a(), cc_mul32_t(),
	cc_div32_at(), cc_div32_ta(), cc_div32_a(), cc_div32_t(),
	cc_waitint(), cc_waitevent(),
	cc_br(), cc_beq(), cc_bne(), cc_bgt(), cc_bge(),
        cc_blt(), cc_ble(), cc_bbc(), cc_bbs(),
        kl_stop(),
	kl_read16(),  kl_read24(), kl_write16(), kl_write24(), kl_ndt(),
	kl_qs16(), kl_qs24(), kl_as16(), kl_as24(), kl_igq16(), kl_igq24();

/* for KEK-list processing */  
int (*instruction_call)();
int (*instruction[0x80])() = {
/* 0 */			cc_stop,
/* 1,2,3,4 */		cc_read16,  cc_read24,  cc_read16_a,  cc_read24_a,
/* 5,6,7,8 */		cc_write16, cc_write24, cc_write16_a, cc_write24_a,
/* 9,A,B,C */		cc_ndt, cc_setcrate_a, cc_setcrate, cc_genz,
/* D,E,F,10,11 */	cc_genc, cc_remi, cc_seti, cc_enbint, cc_dsbint,
/* 12 */		cc_qwait,
/* 13,14,15,16 */	cc_qs16, cc_qs24, cc_qs16_a, cc_qs24_a,
/* 17,18,19,1A */	cc_as16, cc_as24, cc_as16_a, cc_as24_a,
/* 1B,1C,1D,1E */	cc_igq16, cc_igq24, cc_igq16_a, cc_igq24_a,
/* 1F - 22 */		cc_stop, cc_stop, cc_stop, cc_stop,
/* 23,24,25,26 */	cc_move_as, cc_move_at, cc_move_sa, cc_move_st,
/* 27,28 */		cc_move_ta, cc_move_ts,
/* 29,2A,2B,2C */	cc_move32_a, cc_move32_s, cc_move32_t, cc_move16_a_ptr,
/* 2D,2E,2F */		cc_move32_a_ptr, cc_move16_ptr, cc_move32_ptr,
/* 30,31,32 */		cc_move16_a_head, cc_move16_head_a, cc_move16_head,
/* 33,34,35 */		cc_load_ptr, cc_load_remain, cc_load_timer, 
/* 36,37 */		cc_load_count, cc_load_qx,
/* 38,39,3A */		cc_head, cc_exchg_at, cc_swap_a,
/* 3B,3C,3D,3E */	cc_bis32_at, cc_bis32_ta, cc_bis32_a, cc_bis32_t,
/* 3F,40,41,42 */	cc_bic32_at, cc_bic32_ta, cc_bic32_a, cc_bic32_t,
/* 43,44,45,46 */	cc_add32_at, cc_add32_ta, cc_add32_a, cc_add32_t,
/* 47,48,49,4A */	cc_sub32_at, cc_sub32_ta, cc_sub32_a, cc_sub32_t,
/* 4B,4C,4D,4E */	cc_mul32_at, cc_mul32_ta, cc_mul32_a, cc_mul32_t,
/* 4F,50,51,52 */	cc_div32_at, cc_div32_ta, cc_div32_a, cc_div32_t,
/* 53,54 */		cc_waitint, cc_waitevent,
/* 55,56,57,58,59 */	cc_br, cc_beq, cc_bne, cc_bgt, cc_bge,
/* 5A,5B,5C,5D */	cc_blt,cc_ble, cc_bbc, cc_bbs,
/* 5E,5F */		cc_stop, cc_stop,

/* for K2917 list processing */
/* 0x60 + */
/* 0 */			kl_stop,
/* 1,2,3,4 */		kl_read16,  kl_read24,  kl_stop, kl_stop,
/* 5,6,7,8 */		kl_write16, kl_write24, kl_stop, kl_stop,
/* 9,A,B,C */		kl_ndt, kl_stop, kl_stop, kl_stop,
/* D,E,F,10,11 */	kl_stop, kl_stop, kl_stop, kl_stop, kl_stop,
/* 12 */		kl_stop,
/* 13,14,15,16 */	kl_qs16, kl_qs24, kl_stop, kl_stop,
/* 17,18,19,1A */	kl_as16, kl_as24, kl_stop, kl_stop,
/* 1B,1C,1D,1E */	kl_igq16, kl_igq24, kl_stop, kl_stop

};

/******************************************************
 * Internal data areas                                *
 ******************************************************/
static struct cc_iosb iosb;
static struct cc_message message;
static u_short data_area[CC_LIST_MAX_DATA];
static u_short list_area[CC_LIST_MAX_LENGTH];
static struct buf ccbuf;
static struct klist klist[CC_KLIST_MAX_DIR];
static int ptr_dir;
static u_short ptr_cma, now_cma, now_wc;
static int cc_sys_status;          /* temporally system status storage */

union swap {
    long longs;
    short shorts[2];
} swap;

/******************************************************
 * ccprobe                                            *
 ******************************************************/
ccprobe(reg, unit)
caddr_t reg;
int unit;
{
    register struct K_REG *k;
    register struct cc_device *cc = &ccdevice[0];

    k = (struct K_REG *)reg;

    cc->max_branch = CC_K_MAX_BRANCH;
    cc->cur_branch = 0;
    cc->cur_crate = 0;
    /* check K2917 */
    if (peek((u_short *)&k->csr) == -1)
        return 0;
    poke((u_short *)&k->csr,  (u_short)RST);

    printf("%s",CC_START_MESSAGE);
    return (sizeof(struct K_REG) * cc->max_branch);
}

/******************************************************
 * ccopen                                             *
 ******************************************************/
ccopen(dev, flags)
dev_t dev;
int flags;
{
    register int unit = CCUNIT(dev);
    register struct cc_device *cc = &ccdevice[0];
    register struct mb_device *md = ccdinfo[0];
    void bzero();

    if (unit >= NCC || !md || md->md_alive == 0)
        return (ENXIO);
    if (cc->cc_busy == CC_BUSY)
        return (EBUSY);


    /* initialize ccdevice */
    bzero(cc, sizeof(struct cc_device));
    cc->md = md;
    cc->k = (struct K_REG *)md->md_addr;
    cc->max_branch = CC_K_MAX_BRANCH;

    /* set busy flag */
    cc->cc_busy = CC_BUSY;

    /* set interrupt registers of K2917 */
    cc->k->lamc = (u_short)(INT_AUTO_CLEAR | (md->md_intpri & 7));
    cc->k->donc = (u_short)(INT_AUTO_CLEAR | (md->md_intpri & 7));
    cc->k->empc = (u_short)(INT_AUTO_CLEAR | (md->md_intpri & 7));
    cc->k->aboc = (u_short)(INT_AUTO_CLEAR | (md->md_intpri & 7));
    cc->k->lamv = (u_short)md->md_intr->v_vec;
    cc->k->donv = (u_short)md->md_intr->v_vec;
    cc->k->empv = (u_short)md->md_intr->v_vec;
    cc->k->abov = (u_short)md->md_intr->v_vec;

    return 0;
}

/******************************************************
 * ccclose                                            *
 ******************************************************/
ccclose(dev, flags)
dev_t dev;
int flags;
{
    register int unit = CCUNIT(dev);
    register struct cc_device *cc = &ccdevice[0];

    if (unit >= NCC)
        return ENXIO;
    /* clear busy flag */
    cc->cc_busy = 0;
    if (cc->execution != 0) {
	cc->execution = 0;
	untimeout(cctimeout, cc);
	uprintf("ccclose: cancel timeout.\n");
    }
    return 0;
}

/******************************************************
 * ccwrite                                            *
 ******************************************************/
ccwrite(dev, uio)
dev_t dev;
struct uio *uio;
{
    register int unit = CCUNIT(dev);
    register struct cc_device *cc = &ccdevice[0];
    register struct iovec *iov = uio->uio_iov;
    register u_short mode, naf;
    register int len, code, step_count;
    struct uio   uio_save, *uios = &uio_save;
    struct iovec iov_save, *iovs = &iov_save;
    int retlen;
    char klname[9];

    /* check re-execution (signal interruption at WAITINT, WAITEVENT) */
    if (cc->execution != 0)
        goto resume_list;

    if (unit >= NCC)
        return ENXIO;

    cc->dev = dev;
    cc->uio = uio;
    cc->ptr_kdata = data_area;
    cc->len_kdata = 0;

/*
    if (uiomove(&message, sizeof(struct cc_message), UIO_WRITE, uio))
        return EFAULT;
*/
    copyin((caddr_t)iov->iov_base, (caddr_t)&message, sizeof(message));

    cc->status = 0;
    cc_sys_status = 0;
    cc->mode = mode = message.mode;
    cc->naf  = naf  = message.naf;

/*
    iovs->iov_base   = iov->iov_base;
    iovs->iov_len    = iov->iov_len;
    uios->uio_iovcnt = uio->uio_iovcnt;
    uios->uio_offset = uio->uio_offset;
    uios->uio_resid  = uio->uio_resid;
*/

    switch (message.command) {

	/*************************
	 * CAMAC single action   *
	 *************************/
      case CC_CMD_DOSINGLE:
	switch (naf & 0x0018) {
	  case 0x0000:                                  /* CAMAC read */
	    camac_s(cc, mode, naf, data_area);
	    if ((mode & BIT16) == 0)
	        copyout((caddr_t)data_area, (caddr_t)message.ptr_data,
			sizeof(u_short) * 2);
	    else
	        copyout((caddr_t)data_area, (caddr_t)message.ptr_data,
			sizeof(u_short));
	    break;
	  case 0x0010:                                  /* CAMAC write */
	    if ((mode & BIT16) == 0)
	        copyin((caddr_t)message.ptr_data, (caddr_t)data_area, 
		       sizeof(u_short) * 2);
	    else
	        copyin((caddr_t)message.ptr_data, (caddr_t)data_area,
		       sizeof(u_short));
	    camac_s(cc, mode, naf, data_area);
	    break;
	  default:                                      /* NDT */
	    camac_s(cc, mode, naf, data_area);
	    break;
	}
	copyout((caddr_t)((u_short *)&cc->camac_qx),
		(caddr_t)message.ptr_qx, 2);
        break;

	/*************************
	 * CAMAC block action    *
	 *************************/
      case CC_CMD_DOBLOCK:

	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;

	cc->ptr_udata = (u_short *)iov->iov_base;
	cc->len_udata = 0;
	cc->len_udata_t = len = iov->iov_len / sizeof(u_short);
	if ((mode & BIT16) == 0)
	    len /= 2;

	switch (naf & 0x0018) {
	  case 0x0000:                                  /* CAMAC read */
	    camac_b(cc, mode, naf, len, &retlen);
	    break;
	  case 0x0010:                                  /* CAMAC write */
	    camac_b(cc, mode, naf, len, &retlen);
	    break;
	  default:                                      /* NDT */
	    camac_b(cc, mode, naf, len, &retlen);
	    break;
	}
	iosb.status = cc->status;
	iosb.ret_length = retlen;
	iosb.s_reg = cc->s_reg;
	iosb.devinfo = cc->camac_qx + (cc->camac_qx << 16);
	copyout((caddr_t)&iosb, (caddr_t)message.ptr_iosb,
		sizeof(struct cc_iosb));
        break;

	/*************************
	 * CAMAC list processing *
	 *************************/
      case CC_CMD_LOADLIST:
      case CC_CMD_LOADDOLIST:

	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	cc->ptr_list = (u_short *)iov->iov_base;
	cc->len_list = iov->iov_len / sizeof(u_short);
	copyin((caddr_t)cc->ptr_list, (caddr_t)list_area, 
	       cc->len_list * sizeof(u_short));
	if (message.command == CC_CMD_LOADLIST)
	    break;

      case CC_CMD_DOLIST:
	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	copyin((caddr_t)message.ptr_iosb, (caddr_t)&iosb,
	       sizeof(struct cc_iosb));
	cc->ptr_udata_s = cc->ptr_udata = (u_short *)iov->iov_base;
	cc->len_udata_t = iov->iov_len / sizeof(u_short);
	cc->len_udata = 0;
	cc->ptr_list = list_area;	/* initialize list pointer */
	cc->s_reg = iosb.s_reg;

	step_count = 0;
/* label for re-execution */
      resume_list:
	while (cc->status == CC_STA_OK &&
	       cc_sys_status == CC_STA_OK &&
	       step_count++ < CC_LIST_MAX_STEP &&
	       cc->ptr_list <= (list_area + cc->len_list)) {
	    code = 0x00FF & cc->ptr_list[0];
	    instruction_call = instruction[code];
	    cc->status = (*instruction_call)(cc);
	    next_instruction(cc);
	}
	if (step_count >= CC_LIST_MAX_STEP)
	    cc->status = CC_STA_LIST_OVERSTEP;
	flush_buf(cc);
	iosb.status = cc->status;
	iosb.ret_length = cc->len_udata;
	iosb.s_reg = cc->s_reg;
	iosb.devinfo = cc->camac_qx + (cc->camac_qx << 16);
	copyout((caddr_t)&iosb, (caddr_t)message.ptr_iosb,
				 sizeof(struct cc_iosb));
	break;

	/*************************
	 * K2917 list processing *
	 *************************/
      case CC_CMD_ADDKLIST:

	copyin((caddr_t)message.klname, (caddr_t)klname, sizeof(klname));
	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	cc->ptr_list = (u_short *)iov->iov_base;
	cc->len_list = iov->iov_len / sizeof(u_short);
	copyin((caddr_t)cc->ptr_list, (caddr_t)list_area,
	       cc->len_list * sizeof(u_short));
	kladd(cc, klname);
	break;

      case CC_CMD_DELKLIST:

	copyin((caddr_t)message.klname, (caddr_t)klname, sizeof(klname));
	kldel(klname);
	break;

      case CC_CMD_EXEKLIST:

	copyin((caddr_t)message.klname, (caddr_t)klname, sizeof(klname));
	copyin((caddr_t)message.ptr_iosb, (caddr_t)&iosb,
	       sizeof(struct cc_iosb));
	iov++;
	uio->uio_iov++;
	uio->uio_iovcnt--;
	cc->ptr_udata_s = cc->ptr_udata = (u_short *)iov->iov_base;
	cc->len_udata_t = iov->iov_len / sizeof(u_short);
	cc->len_udata = 0;
	klexe(cc, klname);
	iosb.status = cc->status;
	iosb.ret_length = cc->retlen;
	iosb.s_reg = cc->s_reg;
	iosb.devinfo = cc->camac_qx + (cc->camac_qx << 16);
	copyout((caddr_t)&iosb, (caddr_t)message.ptr_iosb,
				 sizeof(struct cc_iosb));
	break;

      case CC_CMD_SAVEKLIST:

	klsave(cc, data_area);
	copyout((caddr_t)data_area,(caddr_t)message.ptr_iosb,
		(list_area[0] + 1) * 2);
	break;

      case CC_CMD_LOADKLIST:
	copyin((caddr_t)(u_short *)message.ptr_iosb, (caddr_t)data_area, 2);
	copyin((caddr_t)((u_short *)message.ptr_iosb + 1),
	       (caddr_t)&data_area[1], data_area[0] * 2);

	klload(cc, data_area);
	break;

      default:
	break;
    }
/*	
    iov->iov_base   = iovs->iov_base;
    iov->iov_len    = iovs->iov_len;
    uio->uio_iovcnt = uios->uio_iovcnt;
    uio->uio_offset = uios->uio_offset;
    uio->uio_resid  = uios->uio_resid;
*/
    return cc_sys_status;
}

/******************************************************
 * cctimeout                                          *
 ******************************************************/
cctimeout(cc)
struct cc_device *cc;
{
    cc->interrupt |= CC_INT_TIMEOUT;
    wakeup((caddr_t)cc);
    return 0;
}

/******************************************************
 * ccintr                                             *
 ******************************************************/
ccintr(unit)
int unit;
{
    register struct cc_device *cc = &ccdevice[0];
    register struct mb_device *md = ccdinfo[0];
    register struct K_REG *k;

    k = (struct K_REG *)md->md_addr;

    if ((k->csr & LAM) != 0)
	cc->interrupt |= CC_INT_LAM;
    if ((k->csr & DONE) != 0)
	cc->interrupt |= CC_INT_DONE;
    if ((k->empc & INT_ENABLE) == 0)
	cc->interrupt |= CC_INT_EMPTY;
    if ((k->aboc & INT_ENABLE) == 0)
	cc->interrupt |= CC_INT_ABORT;

    untimeout(cctimeout, cc);
    wakeup((caddr_t)cc);
}

/******************************************************
 * ccioctl                                            *
 ******************************************************/
ccioctl(dev, cmd, data, flag)
dev_t dev;
int cmd;
u_short *data;
int flag;
{
    register struct cc_device *cc = &ccdevice[0];
    register struct mb_device *md = ccdinfo[0];
    register struct K_REG *k;
    u_short sdat;
    int     idat;
    int s;

    k = (struct K_REG *)md->md_addr;

    switch (cmd) {
      case CCIOC_CHK_BRANCH:
	*(int *)data = cc->max_branch;
	break;
      case CCIOC_SET_BRANCH:
	if (*(int *)data >= 0 && *(int *)data < cc->max_branch)
	    cc->cur_branch = *(int *)data;
	break;
      case CCIOC_SET_CRATE:
	if (*(int *)data >= 0 && *(int *)data < MAX_CRATE)
	    cc->cur_crate = *(int *)data;
	break;
      case CCIOC_WAIT_LAM :
	if (cc->execution == 0) {        /* not re-execution */
	    cc->interrupt = 0;
	    k->lamc = INT_AUTO_CLEAR | INT_ENABLE | (md->md_intpri & 7);
	}
	/* lock out clock */
	s = spl5();
	if (cc->interrupt & CC_INT_LAM) {
	    cc->interrupt &= ~CC_INT_LAM;
	    cc->execution = 0;
	    return 0;
	}
	if (cc->execution == 0) {        /* not re-execution */
	    timeout(cctimeout, cc, *(int *)data * hz);
	    cc->execution = 1;
	}
	else
  	  uprintf("cwlam: re-exec...\n");
	sleep((caddr_t)cc, CCLPRI);
	splx(s);
	cc->execution = 0;

	if( cc->interrupt & CC_INT_LAM ) {
	    cc->interrupt &= ~CC_INT_LAM;
	} else if( cc->interrupt & CC_INT_TIMEOUT ) {
	    cc->interrupt &= ~CC_INT_TIMEOUT;
	    return CC_STA_SINGLE_TIMEOUT;   /* set value to "errno" */
	}
	break;

      case CCIOC_ENABLE_LAM :
	camac_s(cc, BIT24, NAF(30, 13, 17), (u_short *)data); /* write mask */
	camac_s(cc, BIT16, NAF(30,  0,  1), &sdat);
	sdat |= 0x100;
	camac_s(cc, BIT16, NAF(30,  0, 17), &sdat);
	break;
      case CCIOC_DISABLE_LAM :
	k->lamc = INT_AUTO_CLEAR | (md->md_intpri & 7);
	idat = 0;
	camac_s(cc, BIT24, NAF(30, 13, 17), (u_short *)&idat); /* write mask */
	camac_s(cc, BIT16, NAF(30,  0,  1), &sdat);
	sdat &= ~0x100;
	camac_s(cc, BIT16, NAF(30,  0, 17), &sdat);
	break;
      case CCIOC_REG_DUMP :
/*
        data[0] = k->cser;
        data[1] = k->docr;
	data[2] = k->sccr;
	data[3] = k->mtc;
	data[4] = k->machi;
	data[5] = k->maclo;
	data[6] = k->lamc;
	data[7] = k->donc;
	data[8] = k->empc;
	data[9] = k->aboc;
	data[10] = k->lamv;
	data[11] = k->donv;
	data[12] = k->empv;
	data[13] = k->abov;
	data[14] = k->amr;
	data[15] = k->cma;
	data[16] = k->cwc;
	data[17] = k->srr;
	data[18] = k->csr;
*/
	k->csr = RST;
	break;
      case CCIOC_RESET :
	k->csr = RST;
	break;
      case CCIOC_READ_STATUS :
/*
        data[0] = k->cser;
        data[1] = k->docr;
	data[2] = k->sccr;
	data[3] = k->mtc;
	data[4] = k->machi;
	data[5] = k->maclo;
	data[6] = k->lamc;
	data[7] = k->donc;
	data[8] = k->empc;
	data[9] = k->aboc;
	data[10] = k->lamv;
	data[11] = k->donv;
	data[12] = k->empv;
	data[13] = k->abov;
	data[14] = k->amr;
	data[15] = k->cma;
	data[16] = k->cwc;
	data[17] = k->srr;
	data[18] = k->csr;
*/
	k->csr = RST;
	break;

      case CCIOC_KINIT :
	ptr_dir = 0;
	ptr_cma = CC_KLIST_CMAINIT;
	break;


      default :
	break;
    }

    return 0;
}
